"""Multiprocessing worker."""

from argparse import Namespace
from datetime import datetime
from logging import INFO, Logger, getLogger
from multiprocessing import Process, Queue
from queue import Empty
from signal import SIGUSR1, SIGUSR2, signal
from typing import Any, Iterable, Iterator, Sequence, Type

from setproctitle import setproctitle

from homeinfotools.exceptions import SSHConnectionError
from homeinfotools.logging import syslogger


__all__ = ['BaseWorker', 'multiprocess']


class BaseWorker:
    """Stored args and manager to process systems."""

    __slots__ = ('index', 'systems', 'results', 'running', 'current_system')

    def __init__(self, index: int, systems: Queue, results: Queue):
        """Sets the command line arguments."""
        self.index = index
        self.systems = systems
        self.results = results
        self.running = True
        self.current_system = None

    def __call__(self, args: Namespace) -> None:
        """Runs the worker on the given system."""
        setproctitle(self.name)
        signal(SIGUSR1, self.signal)
        signal(SIGUSR2, self.signal)

        while self.running:
            try:
                self.current_system = system = self.systems.get(timeout=1)
            except Empty:
                self.logger.info('Finished')
                return

            result = self.process_system(system, args)
            self.results.put_nowait((system, result))

        self.logger.info('Aborted')

    @property
    def info(self) -> str:
        """Returns information about the state of the worker."""
        if self.current_system is None:
            return 'idle'

        return f'Processing system #{self.current_system}'

    @property
    def logger(self) -> Logger:
        """Returns the worker's logger."""
        logger = getLogger(self.name)
        logger.setLevel(INFO)
        return logger

    @property
    def name(self) -> str:
        """Returns the worker's name."""
        return f'hidsltools-worker-{self.index}'

    def signal(self, signal_number: int, _: Any) -> None:
        """Handles the given signal."""
        if signal_number == SIGUSR1:
            self.logger.info(self.info)
        elif signal_number == SIGUSR2:
            self.running = False
        else:
            self.logger.error('Received invalid signal: %i', signal_number)

    def process_system(self, system: int, args: Namespace) -> dict:
        """Processes a single system."""
        result = {'start': (start := datetime.now()).isoformat()}

        try:
            result['result'] = self.run(system, args)
        except SSHConnectionError:
            syslogger(system).error('Could not establish SSH connection.')
            result['success'] = False
        else:
            result['success'] = True

        result['end'] = (end := datetime.now()).isoformat()
        result['duration'] = str(end - start)
        return result

    @staticmethod
    def run(system: int, args: Namespace) -> dict:
        """Runs the respective processes."""
        raise NotImplementedError()


def multiprocess(
        worker_cls: Type[BaseWorker],
        systems: list[int],
        processes: int,
        args: Namespace
) -> dict:
    """Spawns workers and waits for them to finish."""

    results = Queue()
    wait_for_processes(
        spawn_workers(
            worker_cls,
            processes,
            sequence_to_queue(systems),
            results,
            args
        )
    )
    return dict(iter_queue(results))


def sequence_to_queue(sequence: Sequence[Any]) -> Queue:
    """Returns a queue with items from the given sequence."""

    queue = Queue(len(sequence))

    for item in sequence:
        queue.put(item)

    return queue


def spawn_workers(
        worker_cls: Type[BaseWorker],
        amount: int,
        systems: Queue,
        results: Queue,
        args: Namespace
) -> Iterator[Process]:
    """Spawns worker processes."""

    for index in range(amount):
        worker = worker_cls(index, systems, results)
        process = Process(target=worker, args=(args,))
        process.start()
        yield process


def wait_for_processes(processes: Iterable[Process]) -> None:
    """Wait for the given processes."""

    processes = list(processes)

    try:
        for process in processes:
            process.join()
    except KeyboardInterrupt:
        for process in processes:
            process.kill()


def iter_queue(queue: Queue) -> Iterator[Any]:
    """Returns queue elements as dict."""

    while not queue.empty():
        yield queue.get_nowait()



"""Operating system-specific implementations."""

from os import getenv, name
from pathlib import Path


__all__ = ['SSH', 'RSYNC', 'CACHE_DIR']


if name == 'posix':
    SSH = '/usr/bin/ssh'
    RSYNC = '/usr/bin/rsync'
    CACHE_DIR = Path.home() / '.cache'
elif name == 'nt':
    SSH = 'ssh'
    RSYNC = 'rsync'
    CACHE_DIR = Path(getenv('TEMP') or getenv('TMP'))
else:
    raise NotImplementedError('Unsupported operating system.')



"""Argument parsing."""

from argparse import ArgumentParser, Namespace
from pathlib import Path

from homeinfotools.os import CACHE_DIR


__all__ = ['get_args']


CACHE_FILE = CACHE_DIR / 'sysquery.cache'


def get_args() -> Namespace:
    """Returns parsed CLI arguments."""

    parser = ArgumentParser(description='Retrieve and filter systems.')
    parser.add_argument('-U', '--user', metavar='account',
                        help='the HIS user account')
    parser.add_argument('-i', '--id', nargs='+', type=int, metavar='id',
                        help='filter by system IDs')
    parser.add_argument('-o', '--os', nargs='+', metavar='operating system',
                        help='filter by operating systems')
    parser.add_argument('--sn', nargs='+', metavar='serial number',
                        help='filter by serial numbers')
    parser.add_argument('-D', '--deployment', nargs='+', type=int,
                        metavar='deployment', help='filter by deployments')
    parser.add_argument('-C', '--customer', nargs='+', type=str.casefold,
                        metavar='customer', help='filter by customers')
    parser.add_argument('-t', '--type', nargs='+', metavar='type',
                        help='filter by types')
    parser.add_argument('-s', '--street', nargs='+', type=str.casefold,
                        metavar='street', help='filter by streets')
    parser.add_argument('-H', '--house-number', nargs='+', type=str.casefold,
                        metavar='house number', help='filter by house numbers')
    parser.add_argument('-z', '--zip-code', nargs='+', type=str.casefold,
                        metavar='zip code', help='filter by zip codes')
    parser.add_argument('-c', '--city', nargs='+', type=str.casefold,
                        metavar='city', help='filter by cities')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='enable verbose mode')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='enable debug mode')
    parser.add_argument('-f', '--refresh', action='store_true',
                        help='force refreshing of cache')
    parser.add_argument('--cache-file', type=Path, default=CACHE_FILE,
                        metavar='file', help='use the given file as cache')
    parser.add_argument('--cache-time', type=int, default=24, metavar='hours',
                        help='sets the caching time in hours')
    return parser.parse_args()



"""Common functions."""

from argparse import Namespace

from homeinfotools.his import DownloadError, HISSession
from homeinfotools.logging import LOGGER
from homeinfotools.termgr import FINALIZE_URL, VPN_URL


__all__ = ['configure']


def get_vpn_data(session: HISSession, system: int, windows: bool) -> bytes:
    """Retrieves OpenVPN data for the respective system."""

    json = {'system': system, 'windows': windows}
    response = session.post(VPN_URL, json=json)

    if response.status_code != 200:
        raise DownloadError(response)

    return response.content


def finalize_system(session: HISSession, system: int, *, serial_number: str,
                    operating_system: str, model: str, pubkey: str) -> bool:
    """Finalizes the system."""

    json = {'system': system}

    if serial_number is not None:
        json['sn'] = serial_number

    if operating_system is not None:
        json['os'] = operating_system

    if model is not None:
        json['model'] = model

    if pubkey is not None:
        json['wg_pubkey'] = pubkey

    response = session.post(FINALIZE_URL, json=json)

    if response.status_code != 200:
        LOGGER.error('Could not finalize system.')
        LOGGER.debug(response.content)
        return False

    return True


def configure(user: str, passwd: str, args: Namespace) -> bytes:
    """Retrieves OpenVPN data for the respective system."""

    with HISSession(user, passwd) as session:
        vpn_data = get_vpn_data(session, args.system, args.windows)
        finalize_system(
            session, args.system, serial_number=args.serial_number,
            operating_system=args.operating_system, model=args.model,
            pubkey=args.pubkey)
        return vpn_data



"""Argument parsing."""

from argparse import ArgumentParser, Namespace
from multiprocessing import cpu_count
from pathlib import Path


__all__ = ['get_args']


def get_args() -> Namespace:
    """Returns parsed CLI arguments."""

    parser = ArgumentParser(description='Transfer files from and to systems.')
    parser.add_argument(
        'system', type=int, nargs='+', help='systems to upgrade'
    )
    parser.add_argument(
        '-S', '--send', action='store_true',
        help='send a file to the system(s)'
    )
    parser.add_argument(
        '-R', '--retrieve', action='store_true',
        help='retrieve a file from the system(s)'
    )
    parser.add_argument('src', type=Path, help='the source file')
    parser.add_argument('dst', type=Path, help='the destination file')
    parser.add_argument(
        '-d', '--debug', action='store_true', help='enable debug logging'
    )
    parser.add_argument(
        '-p', '--processes', type=int, metavar='n', default=cpu_count(),
        help='amount of parallel processes'
    )
    parser.add_argument(
        '-s', '--shuffle', action='store_true', help='shuffle systems'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='enable verbose logging'
    )
    args = parser.parse_args()

    if sum((args.send, args.retrieve)) != 1:
        parser.error('Must specify either send xor receive.')

    return args



"""Execute commands on a remote system."""

from argparse import Namespace

from homeinfotools.exceptions import SSHConnectionError
from homeinfotools.logging import syslogger
from homeinfotools.functions import completed_process_to_json, execute
from homeinfotools.ssh import ssh


__all__ = ['runcmd']


def runcmd(system: int, args: Namespace) -> bool:
    """Runs commands on a remote system."""

    command = ssh(system, args.execute, no_stdin=args.no_stdin)
    syslogger(system).debug('Running "%s" on system.', args.execute)
    completed_process = execute(command)

    if completed_process.returncode == 255:
        raise SSHConnectionError(completed_process)

    syslogger(system).debug('Returncode: %i', completed_process.returncode)

    if stdout := completed_process.stdout:
        syslogger(system).info(stdout.strip())

    if stderr := completed_process.stderr:
        syslogger(system).warning(stderr.strip())

    if not completed_process.returncode == 0:
        syslogger(system).error('Command failed.')

    return completed_process_to_json(completed_process)



"""Reboots a system."""

from argparse import Namespace

from homeinfotools.exceptions import SSHConnectionError
from homeinfotools.functions import completed_process_to_json, execute
from homeinfotools.logging import syslogger
from homeinfotools.rpc.common import SYSTEMCTL
from homeinfotools.rpc.sudo import sudo
from homeinfotools.ssh import ssh


__all__ = ['reboot']


def reboot(system: int, args: Namespace) -> bool:
    """Reboots a system."""

    command = ssh(system, *sudo(SYSTEMCTL, 'reboot'), no_stdin=args.no_stdin)
    syslogger(system).debug('Rebooting system %i.', system)
    completed_process = execute(command)

    if completed_process.returncode == 0:
        syslogger(system).info('System %i is rebooting.', system)
    elif completed_process.returncode == 1:
        syslogger(system).warning('System %i may be rebooting.', system)
    elif completed_process.returncode == 255:
        raise SSHConnectionError(completed_process)

    return completed_process_to_json(completed_process)



"""Batch sync files."""

from logging import basicConfig
from random import shuffle

from homeinfotools.functions import get_log_level, handle_keyboard_interrupt
from homeinfotools.logging import LOG_FORMAT
from homeinfotools.multiprocessing import multiprocess
from homeinfotools.filetransfer.argparse import get_args
from homeinfotools.filetransfer.worker import Worker


__all__ = ['main']


@handle_keyboard_interrupt
def main() -> None:
    """Runs the script."""

    args = get_args()
    basicConfig(format=LOG_FORMAT, level=get_log_level(args))

    if args.shuffle:
        shuffle(args.systems)

    multiprocess(Worker, args.system, args.processes, args=args)



"""Main script."""

from logging import DEBUG, INFO, WARNING, basicConfig

from homeinfotools.functions import handle_keyboard_interrupt
from homeinfotools.his import ErrorHandler
from homeinfotools.logging import LOG_FORMAT, LOGGER
from homeinfotools.query.argparse import get_args
from homeinfotools.query.functions import filter_systems, get_systems


__all__ = ['main']


@handle_keyboard_interrupt
def main() -> None:
    """Runs the script."""

    args = get_args()
    loglevel = DEBUG if args.debug else INFO if args.verbose else WARNING
    basicConfig(format=LOG_FORMAT, level=loglevel)
    LOGGER.info('Retrieving systems.')

    with ErrorHandler('Error during JSON data retrieval.'):
        systems = get_systems(args)

    for system in filter_systems(systems, args):
        print(system.get('id'))



"""Transfer files."""

from argparse import Namespace
from pathlib import Path

from homeinfotools.exceptions import SSHConnectionError
from homeinfotools.logging import syslogger
from homeinfotools.functions import completed_process_to_json, execute
from homeinfotools.ssh import rsync


__all__ = ['filetransfer']


def send(system: int, src: Path, dst: Path) -> list[str]:
    """Sends a file to the system."""

    return rsync(src, (system, dst))


def retrieve(system: int, src: Path, dst: Path) -> list[str]:
    """Retrieves a file from the system."""

    dst = dst.parent / (dst.stem + f'.{system}' + dst.suffix)
    return rsync((system, src), dst)


def filetransfer(system: int, args: Namespace) -> bool:
    """Runs commands on a remote system."""

    if args.retrieve:
        command = retrieve(system, args.src, args.dst)
    elif args.send:
        command = send(system, args.src, args.dst)

    syslogger(system).debug('Sending "%s" to system.', args.src)
    completed_process = execute(command)

    if completed_process.returncode == 255:
        raise SSHConnectionError(completed_process)

    syslogger(system).debug('Returncode: %i', completed_process.returncode)

    if stdout := completed_process.stdout:
        syslogger(system).info(stdout.strip())

    if stderr := completed_process.stderr:
        syslogger(system).warning(stderr.strip())

    if not completed_process.returncode == 0:
        syslogger(system).error('File transfer failed.')

    return completed_process_to_json(completed_process)



"""Argument parsing."""

from argparse import ArgumentParser, Namespace
from pathlib import Path


__all__ = ['get_args']


def get_args() -> Namespace:
    """Parses the command line arguments."""

    parser = ArgumentParser(description='OpenVPN configuration retrieval.')
    parser.add_argument('system', type=int, help='the system ID')
    parser.add_argument('-d', '--debug', action='store_true',
                        help='enable debug mode')
    parser.add_argument('-f', '--file', type=Path, metavar='filename',
                        help='output file')
    parser.add_argument('-m', '--model', metavar='model', help='Set model')
    parser.add_argument('-o', '--operating-system', metavar='operating_system',
                        help='Set OS')
    parser.add_argument('-p', '--pubkey', metavar='public_key',
                        help='Set WireGuard pubkey')
    parser.add_argument('-s', '--serial-number', metavar='serial_number',
                        help='Set serial number')
    parser.add_argument('-u', '--user', metavar='his_account',
                        help='HIS user name')
    parser.add_argument('-w', '--windows', action='store_true',
                        help='package for MS Windows systems')
    return parser.parse_args()



"""OpenVPN configuration package client."""

from logging import DEBUG, INFO, basicConfig
from sys import stdout

from homeinfotools.functions import handle_keyboard_interrupt
from homeinfotools.his import update_credentials, ErrorHandler
from homeinfotools.logging import LOG_FORMAT
from homeinfotools.vpn.argparse import get_args
from homeinfotools.vpn.functions import configure


__all__ = ['main']


@handle_keyboard_interrupt
def main() -> None:
    """Main script."""

    args = get_args()
    basicConfig(format=LOG_FORMAT, level=DEBUG if args.debug else INFO)
    user, passwd = update_credentials(args.user)

    with ErrorHandler('Error during VPN data retrieval.'):
        tar_file = configure(user, passwd, args)

    if args.file is None:
        stdout.buffer.write(tar_file)
    else:
        with args.file.open('wb') as file:
            file.write(tar_file)



"""Terminal batch updating utility."""

from json import dump
from logging import basicConfig
from random import shuffle

from homeinfotools.functions import get_log_level, handle_keyboard_interrupt
from homeinfotools.logging import LOG_FORMAT
from homeinfotools.multiprocessing import multiprocess
from homeinfotools.rpc.argparse import get_args
from homeinfotools.rpc.worker import Worker


__all__ = ['main']


@handle_keyboard_interrupt
def main() -> None:
    """Runs the script."""

    args = get_args()
    basicConfig(format=LOG_FORMAT, level=get_log_level(args))

    if args.shuffle:
        shuffle(args.systems)

    result = multiprocess(Worker, args.system, args.processes, args=args)

    if args.json is not None:
        with args.json.open('w') as file:
            dump(dict(result), file, indent=2)



"""Argument parsing."""

from argparse import ArgumentParser, Namespace
from multiprocessing import cpu_count
from pathlib import Path


__all__ = ['get_args']


def get_args() -> Namespace:
    """Returns parsed CLI arguments."""

    parser = ArgumentParser(description='Batch upgrade systems.')
    parser.add_argument(
        'system', type=int, nargs='+', help='systems to upgrade'
    )
    parser.add_argument(
        '-S', '--sysupgrade', action='store_true', help='upgrade the systems'
    )
    parser.add_argument(
        '-X', '--execute', metavar='command',
        help='execute the commands on the systems'
    )
    parser.add_argument(
        '-R', '--reboot', action='store_true', help='reboot the systems'
    )
    parser.add_argument(
        '-c', '--cleanup', action='store_true',
        help='cleanup unneeded packages after upgrade'
    )
    parser.add_argument(
        '-d', '--debug', action='store_true', help='enable debug logging'
    )
    parser.add_argument(
        '-i', '--install', default=(), nargs='+', metavar='pkg',
        help='install the given packages'
    )
    parser.add_argument(
        '-j', '--json', type=Path, metavar='file',
        help='log jobs as JSON when done'
    )
    parser.add_argument(
        '-k', '--keyring', action='store_true',
        help='upgrade the keyring before sysupgrade'
    )
    parser.add_argument(
        '-n', '--no-stdin', action='store_true',
        help='make ssh to not read STDIN'
    )
    parser.add_argument(
        '-o', '--overwrite', default=(), nargs='+', metavar='glob',
        help='globs of files to overwrite'
    )
    parser.add_argument(
        '-p', '--processes', type=int, metavar='n', default=cpu_count(),
        help='amount of parallel processes'
    )
    parser.add_argument(
        '-s', '--shuffle', action='store_true', help='shuffle systems'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='enable verbose logging'
    )
    parser.add_argument(
        '-y', '--yes', action='store_true',
        help='assume yes on program queries'
    )
    return parser.parse_args()



"""HIS SSO API."""

from requests import Response


__all__ = ['WebAPIError', 'DownloadError', 'LoginError']


class WebAPIError(Exception):
    """A web API error."""

    def __init__(self, response: Response):
        """Sets the response."""
        super().__init__(response)
        self.response = response

    def __str__(self):
        """Returns the content as string."""
        return self.response.text


class DownloadError(WebAPIError):
    """Indicates an error during data download."""


class LoginError(WebAPIError):
    """Indicates an error during login."""



"""Processing of systems."""

from argparse import Namespace

from homeinfotools.filetransfer.filetransfer import filetransfer
from homeinfotools.multiprocessing import BaseWorker


__all__ = ['Worker']


class Worker(BaseWorker):
    """Stored args and manager to process systems."""

    @staticmethod
    def run(system: int, args: Namespace) -> dict:
        """Runs the worker."""
        return {'rsync': filetransfer(system, args)}



"""Common exceptions."""

from subprocess import CompletedProcess


__all__ = ['RemoteProcessError', 'SSHConnectionError']


class RemoteProcessError(Exception):
    """Error when executing processes on remote systems."""

    def __init__(self, completed_process: CompletedProcess):
        """Creates an exception from the given completed process."""
        super().__init__(completed_process)
        self.completed_process = completed_process

    def __getattr__(self, attribute):
        """Delegates to completed_process."""
        return getattr(self.completed_process, attribute)

    def __str__(self):
        """Returns a string representation of the error."""
        items = []

        if stdout := self.stdout:
            items.append(f'STDOUT: {stdout}')

        if stderr := self.stderr:
            items.append(f'STDERR: {stderr}')

        items.append(f'EXIT_CODE: {self.returncode}')
        return ' / '.join(items)


class SSHConnectionError(RemoteProcessError):
    """Indicates that the SSH could not connect to the system."""



"""HIS SSO API."""

from sys import exit    # pylint: disable=W0622

from homeinfotools.his.exceptions import DownloadError, LoginError
from homeinfotools.logging import LOGGER


__all__ = ['ErrorHandler']


class ErrorHandler:
    """Handles login and download errors."""

    def __init__(self, error_text: str):
        """Sets the download error text."""
        self.error_text = error_text

    def __enter__(self):
        return self

    def __exit__(self, _, value, __):
        """Handles login and download errors."""
        if isinstance(value, LoginError):
            LOGGER.error('Error during login.')
            LOGGER.debug(value)
            exit(2)

        if isinstance(value, DownloadError):
            LOGGER.error(self.error_text)
            LOGGER.debug(value)
            exit(3)



"""HIS SSO API."""

from homeinfotools.his.errorhandler import ErrorHandler
from homeinfotools.his.exceptions import DownloadError, LoginError
from homeinfotools.his.functions import update_credentials
from homeinfotools.his.session import HISSession


__all__ = [
    'DownloadError',
    'LoginError',
    'ErrorHandler',
    'HISSession',
    'update_credentials'
]



"""Common constants."""

from urllib.parse import urljoin


__all__ = ['DEPLOYMENTS_URL', 'FINALIZE_URL', 'SYSTEMS_URL', 'VPN_URL']


BASE_URL = 'https://termgr.homeinfo.de'
DEPLOYMENTS_URL = urljoin(BASE_URL, '/list/deployments')
FINALIZE_URL = urljoin(BASE_URL, '/setup/finalize')
SYSTEMS_URL = urljoin(BASE_URL, '/list/systems')
VPN_URL = urljoin(BASE_URL, '/setup/openvpn')



"""System upgrade."""

from argparse import Namespace
from subprocess import CompletedProcess

from homeinfotools.exceptions import SSHConnectionError
from homeinfotools.functions import completed_process_to_json, execute
from homeinfotools.logging import syslogger
from homeinfotools.rpc.common import PACMAN
from homeinfotools.rpc.exceptions import PacmanError
from homeinfotools.rpc.exceptions import SystemIOError
from homeinfotools.rpc.exceptions import UnknownError
from homeinfotools.rpc.sudo import sudo
from homeinfotools.ssh import ssh


__all__ = ['sysupgrade']


PACMAN_WRAPPER = '%s; EXIT_CODE=$?; wait; exit ${EXIT_CODE};'


def lograise(system: int, message: str,
             completed_process: CompletedProcess) -> None:
    """Issues a warning message and raises an exception."""

    if completed_process.returncode == 255:
        raise SSHConnectionError(completed_process)

    # Do not warn on SSH connection errors.
    syslogger(system).warning(message)

    if completed_process.returncode == 126:
        raise SystemIOError(completed_process)

    if completed_process.returncode == 1:
        raise PacmanError(completed_process)

    raise UnknownError(completed_process)


def upgrade_keyring(system: int, args: Namespace) -> CompletedProcess:
    """Upgrades the archlinux-keyring on that system."""

    command = [
        PACMAN, '-Sy', 'archlinux-keyring', '--needed', '--noconfirm',
        '--disable-download-timeout'
    ]
    command = sudo(*command)
    command = ssh(system, *command, no_stdin=args.no_stdin)
    syslogger(system).debug('Executing command: %s', command)
    return execute(command)


def upgrade_system(system: int, args: Namespace) -> CompletedProcess:
    """Upgrades the system."""

    command = [PACMAN, '-Syu', '--needed', '--disable-download-timeout']

    for package in args.install:
        command.append(package)

    for glob in args.overwrite:
        command.append('--overwrite')
        command.append(glob)

    if not args.yes:
        command.append('--noconfirm')

    command = ' '.join(sudo(*command))

    if args.yes:
        command = f'yes | {command}'

    command = ssh(system, PACMAN_WRAPPER % command, no_stdin=args.no_stdin)
    syslogger(system).debug('Executing command: %s', command)
    return execute(command)


def cleanup_system(system: int, args: Namespace) -> CompletedProcess:
    """Cleans up the system."""

    command = [PACMAN, '-Rncs', '$(pacman -Qmq; pacman -Qdtq)']

    if not args.yes:
        command.append('--noconfirm')

    command = ' '.join(sudo(*command))

    if args.yes:
        command = f'yes | {command}'

    command = ssh(system, PACMAN_WRAPPER % command, no_stdin=args.no_stdin)
    syslogger(system).debug('Executing command: %s', command)
    return execute(command)


def upgrade(system: int, args: Namespace) -> dict:
    """Upgrade process function."""

    syslogger(system).info('Upgrading system.')
    result = {}

    if args.keyring:
        completed_process = upgrade_keyring(system, args=args)
        result['keyring'] = completed_process_to_json(completed_process)

        if completed_process.returncode != 0:
            lograise(system, 'Could not update keyring.', completed_process)

    completed_process = upgrade_system(system, args=args)
    result['sysupgrade'] = completed_process_to_json(completed_process)

    if completed_process.returncode != 0:
        lograise(system, 'Could not upgrade system.', completed_process)

    if args.cleanup:
        completed_process = cleanup_system(system, args=args)
        result['pkgcleanup'] = completed_process_to_json(completed_process)

        if completed_process.returncode not in {0, 1}:
            lograise(system, 'Could not clean up system.', completed_process)

    return result


def sysupgrade(system: int, args: Namespace) -> bool:
    """Upgrated the respective system."""

    try:
        return upgrade(system, args)
    except SystemIOError as error:
        syslogger(system).error('I/O error.')
        syslogger(system).debug('%s', error)
        return completed_process_to_json(error.completed_process)
    except PacmanError as error:
        syslogger(system).error('Pacman error.')
        syslogger(system).debug('%s', error)
        return completed_process_to_json(error.completed_process)
    except UnknownError as error:
        syslogger(system).error('Unknown error.')
        syslogger(system).debug('%s', error)
        return completed_process_to_json(error.completed_process)



"""Common constants."""

from logging import getLogger
from pathlib import Path
from sys import argv


__all__ = ['LOG_FORMAT', 'LOGGER', 'syslogger']


LOG_FORMAT = '[%(levelname)s] %(name)s: %(message)s'
LOGGER = getLogger(Path(argv[0]).name)


def syslogger(system: int):
    """Returns a sub-logger for the given system."""

    return LOGGER.getChild(str(system))



#! /usr/bin/env python3
"""Installation script."""

from setuptools import setup


with open('README.md', encoding='utf-8') as readme:
    README = readme.read()


setup(
    name='homeinfotools',
    use_scm_version=True,
    setup_requires=['requests', 'setuptools_scm'],
    author='HOMEINFO - Digitale Informationssysteme GmbH',
    author_email='info@homeinfo.de',
    maintainer='Richard Neumann',
    maintainer_email='r.neumann@homeinfo.de',
    python_requires='>=3.8',
    install_requires=[
        'setproctitle'
    ],
    packages=[
        'homeinfotools',
        'homeinfotools.filetransfer',
        'homeinfotools.his',
        'homeinfotools.query',
        'homeinfotools.rpc',
        'homeinfotools.vpn'
    ],
    entry_points={
        'console_scripts': [
            'sysquery = homeinfotools.query.main:main',
            'sysrpc = homeinfotools.rpc.main:main',
            'sysrsync = homeinfotools.filetransfer.main:main',
            'sysvpn = homeinfotools.vpn.main:main',
        ],
    },
    url='https://github.com/homeinfogmbh/homeinfotools',
    license='GPLv3',
    description='Tools to manage HOMEINFO digital signge systems.',
    long_description=README,
    long_description_content_type="text/markdown",
    keywords='python HOMEINFO systems client'
)



"""Query systems."""

from argparse import Namespace
from datetime import datetime, timedelta
from functools import partial
from json import dump, load
from typing import Iterable

from homeinfotools.his import update_credentials, HISSession
from homeinfotools.logging import LOGGER
from homeinfotools.termgr import SYSTEMS_URL


__all__ = ['get_systems', 'filter_systems']


def query_systems(account: str, passwd: str) -> list:
    """Query systems."""

    LOGGER.debug('Querying systems.')

    with HISSession(account, passwd) as session:
        return session.get_json(SYSTEMS_URL)


def cache_systems(systems: list, args: Namespace) -> list:
    """Caches the systems and returns them."""

    cache = {'timestamp': datetime.now().isoformat(), 'systems': systems}

    with args.cache_file.open('w') as file:
        dump(cache, file)

    return systems


def systems_from_cache(args: Namespace) -> list:
    """Returns cached systems."""

    if not args.cache_file.exists():
        LOGGER.info('Initializing cache.')
        systems = query_systems(*update_credentials(args.user))
        return cache_systems(systems, args)

    LOGGER.debug('Loading cache.')

    with args.cache_file.open('r') as file:
        cache = load(file)

    if timestamp := cache.get('timestamp'):
        timestamp = datetime.fromisoformat(timestamp)

        if timestamp + timedelta(hours=args.cache_time) > datetime.now():
            return cache['systems']

        LOGGER.info('Cache has expired.')
    else:
        LOGGER.warning('Corrupted cache.')

    return cache_systems(query_systems(*update_credentials(args.user)), args)


def get_systems(args: Namespace) -> list:
    """Returns systems."""

    if args.refresh:
        systems = query_systems(*update_credentials(args.user))
        return cache_systems(systems, args)

    return systems_from_cache(args)


def substr_ic_in(string: str, haystack: Iterable[str]) -> bool:
    """Checks whether the string is a substring of
    any strings in the iterable, ignoring the case.
    """

    if not string:
        return False

    string = string.casefold()
    return any(substring in string for substring in haystack)


# pylint:disable=R0911,R0912
def match_system(system: dict, *, args: Namespace) -> bool:
    """Matches the system to the filters."""

    if args.id and system.get('id') not in args.id:
        return False

    if args.os and system.get('operatingSystem') not in args.os:
        return False

    if args.sn and system.get('serialNumber') not in args.sn:
        return False

    deployment = system.get('deployment') or {}

    if args.deployment and deployment.get('id') not in args.deployment:
        return False

    if args.customer:
        customer = deployment.get('customer') or {}
        company = customer.get('company') or {}

        if cid := customer.get('id'):
            match = str(cid) in args.customer
        else:
            match = False

        match = match or substr_ic_in(company.get('name'), args.customer)
        match = match or substr_ic_in(
            company.get('abbreviation'), args.customer)

        if not match:
            return False

    if args.type and deployment.get('type') not in args.type:
        return False

    address = deployment.get('address') or {}

    if args.street and not substr_ic_in(address.get('street'), args.street):
        return False

    if args.house_number and not substr_ic_in(
            address.get('houseNumber'), args.house_number):
        return False

    if args.zip_code and not substr_ic_in(
            address.get('zipCode'), args.zip_code):
        return False

    if args.city and not substr_ic_in(address.get('city'), args.city):
        return False

    return True


def filter_systems(systems: list, args: Namespace) -> Iterable[dict]:
    """Filter systems according to the args."""

    return filter(partial(match_system, args=args), systems)



"""Common exceptions."""

from homeinfotools.exceptions import RemoteProcessError


__all__ = ['SystemIOError', 'PacmanError', 'UnknownError']


class SystemIOError(RemoteProcessError):
    """Indicates an I/O error on the remote system."""


class PacmanError(RemoteProcessError):
    """Indicates an error with pacman."""


class UnknownError(RemoteProcessError):
    """Indicates an unknown error."""



"""HIS SSO API."""

from typing import Any

from requests import session

from homeinfotools.his.exceptions import DownloadError, LoginError


__all__ = ['HISSession']


URL = 'https://his.homeinfo.de/session'


class HISSession:
    """A HIS session."""

    def __init__(self, account: str, passwd: str):
        """Sets account name and password."""
        self.account = account
        self.passwd = passwd
        self.session = session()
        self.session_guard = None

    def __enter__(self):
        self.session_guard = self.session.__enter__()
        self.login()
        return self

    def __exit__(self, *args):
        self.session_guard = None
        return self.session.__exit__(*args)

    def __getattr__(self, attr: str) -> Any:
        """Delegates to the session object."""
        return getattr(self.session_guard, attr)

    @property
    def json(self):
        """Returns the login credentials as JSON."""
        return {'account': self.account, 'passwd': self.passwd}

    def login(self) -> bool:
        """Performs a login."""
        if (response := self.post(URL, json=self.json)).status_code != 200:
            raise LoginError(response)

        return True

    def get_json(self, url: str) -> dict:
        """Returns a JSON-ish dict."""
        if (response := self.get(url)).status_code != 200:
            raise DownloadError(response)

        return response.json()



"""HIS SSO API."""

from getpass import getpass
from sys import exit    # pylint: disable=W0622
from typing import Optional

from homeinfotools.logging import LOGGER


__all__ = ['update_credentials']


def update_credentials(account: Optional[str],
                       passwd: Optional[str] = None) -> tuple[str, str]:
    """Reads the credentials for a HIS account."""

    if not account:
        try:
            account = input('User name: ')
        except (EOFError, KeyboardInterrupt):
            print()
            LOGGER.error('Aborted by user.')
            exit(1)

    if not passwd:
        try:
            passwd = getpass('Password: ')
        except (EOFError, KeyboardInterrupt):
            print()
            LOGGER.error('Aborted by user.')
            exit(1)

    return (account, passwd)



"""Processing of systems."""

from argparse import Namespace

from homeinfotools.multiprocessing import BaseWorker
from homeinfotools.rpc.reboot import reboot
from homeinfotools.rpc.runcmd import runcmd
from homeinfotools.rpc.sysupgrade import sysupgrade


__all__ = ['Worker']


class Worker(BaseWorker):
    """Stored args and manager to process systems."""

    @staticmethod
    def run(system: int, args: Namespace) -> dict:
        """Runs the worker."""
        result = {}

        if args.sysupgrade:
            result['sysupgrade'] = sysupgrade(system, args)

        if args.execute:
            result['execute'] = runcmd(system, args)

        if args.reboot:
            result['reboot'] = reboot(system, args)

        return result



"""Common functions."""

from argparse import Namespace
from functools import wraps
from logging import DEBUG, INFO, WARNING
from subprocess import DEVNULL, PIPE, run, CompletedProcess
from typing import Callable, Iterable, Union

from homeinfotools.logging import LOGGER


__all__ = [
    'completed_process_to_json',
    'execute',
    'get_log_level',
    'handle_keyboard_interrupt'
]


def completed_process_to_json(completed_process: CompletedProcess) -> dict:
    """Converts a completed process into a JSON-ish dict."""

    return {
        'args': completed_process.args,
        'returncode': completed_process.returncode,
        'stdout': completed_process.stdout,
        'stderr': completed_process.stderr,
    }


def execute(command: Union[str, Iterable[str]]) -> CompletedProcess:
    """Executes the given command."""

    return run(command, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, text=True,
               check=False)


def get_log_level(args: Namespace) -> int:
    """Returns the set logging level."""

    return DEBUG if args.debug else INFO if args.verbose else WARNING


def handle_keyboard_interrupt(
        function: Callable[..., None]) -> Callable[..., int]:
    """Decorator to run a function with handled keyboard interrupt."""

    @wraps(function)
    def wrapper(*args, **kwargs) -> int:
        try:
            function(*args, **kwargs)
        except KeyboardInterrupt:
            LOGGER.error('Aborted by user.')
            return 1

        return 0

    return wrapper



"""SSH command."""

from pathlib import Path
from typing import Optional, Union

from homeinfotools.os import SSH, RSYNC


__all__ = ['ssh', 'rsync']


HOSTNAME = '{}.terminals.homeinfo.intra'
SSH_OPTIONS = [
    'LogLevel=error',
    'UserKnownHostsFile=/dev/null',
    'StrictHostKeyChecking=no',
    'ConnectTimeout=5'
]
HostPath = Union[Path, tuple[int, Path]]


def ssh(system: Optional[int], *command: str,
        no_stdin: bool = False) -> list[str]:
    """Modifies the specified command to
    run via SSH on the specified system.
    """

    cmd = [SSH]

    if no_stdin:
        cmd.append('-n')

    for option in SSH_OPTIONS:
        cmd.append('-o')
        cmd.append(option)

    if system is not None:
        cmd.append(HOSTNAME.format(system))

    if command:
        cmd.append(' '.join(command))

    return cmd


def get_remote_path(path: HostPath) -> str:
    """Returns a host path."""

    try:
        system, path = path
    except TypeError:
        return path

    return HOSTNAME.format(system) + f':{path}'


def rsync(src: HostPath, dst: HostPath, *,
          all: bool = True,     # pylint: disable=W0622
          update: bool = True, verbose: bool = True) -> list[str]:
    """Returns the respective rsync command."""

    cmd = [RSYNC, '-e', ' '.join(ssh(None))]

    if all:
        cmd.append('-a')

    if update:
        cmd.append('-u')

    if verbose:
        cmd.append('-v')

    return cmd + [get_remote_path(src), get_remote_path(dst)]
